;==============================================================================
;*Author:  Jonathan Weaver, jonw0224@netscape.net
;E-mail Contact: jonw0224@netscape.net
;*Description: A multiply and a divide routine
;*Version: 1.0
;*Date: 2/6/2010
;*Filename: multdiv.inc
;
;Versions:  1.0 	2/6/2010	Created file with header
;
;Copyright (C) 2010 Jonathan Weaver
;
;This program is free software; you can redistribute it and/or modify it under 
;the terms of the GNU General Public License as published by the Free Software 
;Foundation; either version 2 of the License, or (at your option) any later 
;version.
;
;This program is distributed in the hope that it will be useful, but WITHOUT 
;ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
;details.
;
;You should have received a copy of the GNU General Public License along with 
;this program; if not, write to the Free Software Foundation, Inc., 
;51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
;
;==============================================================================

;Variables;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;*multA is used as a multiplier by multBB and divisor by DivideIB.  Declare register in main program.
;external multA

;*multB is used as a multiplier by MultBB and as a quotient by DivideIB.  Declare register in main program.
;external multB

;*prodH is the upper byte of a two byte variable.  It is the product of MultBB and the dividend of DivideIB.  Declare register in main program.
;external prodH

;*prodL is the lower byte of a two byte variable.  It is the product of MultBB and the dividend of DivideIB.  Declare register in main program.
;external prodL

;*cnt is a temporary counter variable used by MultBB and DivideIB.  Declare register in main program.
;external cnt

ifdef _MultBB
;*Multiply Routine.  Returns multA*multB as 16 bit prodH:prodL.  Destroys multB.  Affects WREG and STATUS.  Requires 71 cycles to execute and 14 program words.  Compiles if _MultBB is defined.
;*multA: register to multiply by multB
;*multB: register to multiply by multA
;*prodH: upper byte of product
;*prodL: lower byte of product
;*cnt: internal variable used as a counter
MultBB	;public
	clrf prodH					;clear product
	clrf prodL
	movlw 8
	movwf cnt					;bit counter = 8
	movf multA, w
	bcf STATUS,C        		;Clear the carry bit in the status Reg.
MultBBloop
	  rrf multB, F			;Pick least significant bit off
      btfsc STATUS, C
	      addwf prodH, F 	;Bit was a 1, add multA
      rrf prodH, F			;Change significant bit
      rrf prodL, F
      decfsz cnt, F			;Decrement bit counter
      goto MultBBloop
	retlw   0
endif

ifdef _DivideIB
;*Divide Routine.  Returns prodH:prodL / multA as multB and prodH:prodL MOD multA as prodH.  Affects WREG and STATUS.  Requires 96 Cycles to execute and 16 program words.  Compiles if _DivideIB is defined.
;*multA: one byte divisor.  Expects multA to be less than 128 (positive).
;*prodH: upper byte of a 16 bit dividend.  Expects prodH:prodL to be positive.  Routine destroys prodH.
;*prodL: lower byte of a 16 bit dividend.  Routine destroys prodL.
;*multB: single byte quotient
;*cnt: internal variable used as a counter
DivideIB	;public
	clrf multB				;Initialize multB
	movlw 0x08
	movwf cnt				;Counter is 8
	movf multA, W
	bcf STATUS, C
DivIBLoop
		rlf prodL, F
		rlf prodH, F		;Rotate prodH:prodL
		rlf multB, F		;Keep bit significance sychronized
		subwf prodH, F
		btfss prodH, 7		;if prodH >= 0
			bsf multB,0		;MultA divides in once
		btfsc prodH, 7		;if prodH < 0
			addwf prodH, F	;Restore prodH
		decfsz cnt, F
			goto DivIBLoop
	retlw 0
endif
