;==============================================================================
;*Author:  Jonathan Weaver, jonw0224@aim.com
;E-mail Contact: jonw0224@aim.com
;*Description: Two multiply by literal routines
;*Version: 1.0
;*Date: 2/3/2010
;*Filename: multh.inc
;
;Versions:  1.0		 10/24/2006 - Added and tested MULTL_B_B and MULTL_B_B_SHIFT
;					 03/14/2008 - Added MULTL_B_I and MULTL_B_I_SHIFT
;					 02/03/2010 - Tested and refined MULTL_B_I and 
;								  MULTL_B_I_SHIFT.  Also refined and retested
;								  MULTL_B_B and MULTL_B_B_SHIFT
;
;Copyright (C) 2010 Jonathan Weaver
;
;This program is free software; you can redistribute it and/or modify it under 
;the terms of the GNU General Public License as published by the Free Software 
;Foundation; either version 2 of the License, or (at your option) any later 
;version.
;
;This program is distributed in the hope that it will be useful, but WITHOUT 
;ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
;details.
;
;You should have received a copy of the GNU General Public License along with 
;this program; if not, write to the Free Software Foundation, Inc., 
;51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
;
;==============================================================================


;==============================================================================
; H E L P E R   M A C R O S
;==============================================================================

f equ F
w equ W

;Aid MULTL_B_B in shifting.  
;Shifts reg left by lit, restores restreg to WREG when WREG affected and dorest flag is set.
;Assumes carry bit is clear
MULTL_B_B_SHIFT macro reg, lit, restreg, dorest
    local cnt
    if (lit) < 4                    ;1 -> 2 cycles, 2 -> 4 cycles, 3 -> 6 cycles
        cnt = lit
        while cnt > 0
            rlf reg, f
            cnt--
        endw
;        if (lit) == 1
;            bcf reg, 0
;        else
;            if (lit) == 2
;            	bcf reg, 0
;            	bcf reg, 1
;            else
;            	if (lit) > 0
;            		movlw (0xFF << (lit))
;	            	andwf reg, f
;	            	if dorest == 1
;	            		movf restreg, w
;	            	endif
;	            endif
;	        endif
;        endif
    else
        if (lit) < 6
            if (lit) == 4           ;4 -> 4 cycles
                swapf reg, f
                movlw 0xF0
                andwf reg, f
            else                    ;5 -> 5 cycles
                swapf reg, f
                rlf reg, f
                movlw B'11100000'
                andwf reg, f
            endif
            if dorest == 1
            	movf restreg, w
            endif
        else
            if (lit) == 6           ;6 -> 6 cycles
                rrf reg, f
                rrf reg, f
                rrf reg, f
                movlw B'11000000'
                andwf reg, f
	            if dorest == 1
	            	movf restreg, w
	            endif
            else 
            	if (lit) == 7       ;7 -> 3 cycles
                	rrf reg, f
                	clrf reg
                	rrf reg, f
                else				;8 -> 1 cycle
                	clrf reg
                endif
            endif
        endif
    endif
endm

;Aid MULTL_B_I in shifting.  
;Shifts prodH:prodL left by scount, restores restreg to WREG when WREG affected and dorest flag is set.
;If zeroUpper flag is not set, it copies the value from reg to prodH:prodL and clears the carry bit.
;Otherwise, it assumes the carry bit is clear and prodH:prodL is initialized.
MULTL_B_I_SHIFT macro prodH, prodL, scount, restreg, dorest, zeroUpper
    local cnt
	if (scount) == 0
		if zeroUpper == 0
			clrf prodH
			movf restreg, w
			movwf prodL
			if dorest == 1				;added
				bcf STATUS, C			;added
			endif						;added
		endif
		exitm
	endif
	if (scount) == 1
		if zeroUpper == 0
			clrf prodH
			movf restreg, w
			movwf prodL
			bcf STATUS, C				;added
		endif
		bcf STATUS, C
		rlf prodL, f
		rlf prodH, f
		exitm
	endif
	if (scount) == 2
		if zeroUpper == 0
			clrf prodH
			bcf STATUS, C
			rlf restreg, w
			movwf prodL
			rlf prodH, f
			rlf prodL, f
			rlf prodH, f
			if dorest == 1
				movf restreg, w
			endif
		else
			rlf prodL, f
			rlf prodH, f
			rlf prodL, f
			rlf prodH, f
;			bcf prodL, 0			;removed
;			bcf prodL, 1			;removed
		endif
		exitm
	endif
	if (scount) == 3
		if zeroUpper == 0
			clrf prodH
			bcf STATUS, C
			rlf restreg, w
			movwf prodL
			rlf prodH, f
			rlf prodL, f
			rlf prodH, f
			rlf prodL, f
			rlf prodH, f
		else
			rlf prodL, f
			rlf prodH, f
			rlf prodL, f
			rlf prodH, f
			rlf prodL, f
			rlf prodH, f
;			movlw B'11111000'		;removed
;			andwf prodL, f			;removed
		endif
		if dorest == 1
			movf restreg, w
		endif
		exitm
	endif
	if (scount) < 6
		if zeroUpper == 0		;w = p3, p4		prodH : 0.  prodL: p3, p4
			swapf restreg, w
			movwf prodL
			andlw 0x0f
			movwf prodH
			xorwf prodL, f
		else
			movlw 0x0F           ;prodH: p1, p2.  prodL: p3, p4
	        andwf prodH, f       ;0, p2
	        swapf prodH, f       ;p2, 0
	        swapf prodL, f       ;p4, p3
	        andwf prodL, w       ;0, p3
	        iorwf prodH, f       ;p2, p3
	        xorwf prodL, f       ;p4, 0.  prodH: p2, p3.  prodL: p4, 0
		endif
        if scount == 5
			rlf prodL, f
            rlf prodH, f
;            bcf prodL, 0		;removed
        endif
		if dorest == 1
			movf restreg, w
		endif
		exitm
	endif
    if (scount) == 6
		if zeroUpper == 0
			clrf prodL
			movf restreg, w
			movwf prodH
			bcf STATUS, C
			rrf prodH, f
			rrf prodL, f
			rrf prodH, f
			rrf prodL, f
		else
			rrf prodH, f     ;move prodH[0] into carry
			rrf prodL, w     ;move prodH[0] to prodL and prodL[0] to carry
			clrf prodL       ;set lower byte to zero
			rrf prodL, f     ;make prodL[0] into prodL[7]
			rrf prodH, f     ;move prodH[1] into carry
			movwf prodH      ;prodH = prodH[0]:prodL[7-1]
			rrf prodH, f     ;now prodH = prodH[1]:prodH[0]:prodL[7-2]
			rrf prodL, f     ;now prodL = prodL[1-0]:ZEROS
			if dorest == 1
				movf restreg, w
			endif
		endif
		exitm
	endif
	if (scount) == 7
		if zeroUpper == 0
			clrf prodL
			movf restreg, w
			movwf prodH
			bcf STATUS, C
			rrf prodH, f
			rrf prodL, f
		else
			rrf prodH, f     ;move prodH[0] into carry
			rrf prodL, w	 ;prodH[0]:prodL[7-1], prodL[0] -> C
			movwf prodH
			clrf prodL
			rrf prodL, f	;prodL = prodL[0]:zeros
			if dorest == 1
				movf restreg, w
			endif
		endif
		exitm
	endif
	if (scount) == 8
		if zeroUpper == 0
			clrf prodL
			movf restreg, w
			movwf prodH
		else
			movf prodL, w
			movwf prodH
			clrf prodL
			if dorest == 1
				movf restreg, w
			endif
		endif
	endif
endm

;==============================================================================
; P U B L I C   M A C R O S
;==============================================================================

;------------------------------------------------------------------------------
;*Multiply reg by a literal and store a byte result in prod.  Generates fairly optimized code (11.9 program words/execution cycles on average).  Affects STATUS and WREG.
;*literal: constant to multiply reg by
;*reg: register to multiply by literal
;*prod: register to retain single byte product
MULTL_B_B macro literal, reg, prod 		;public
variable count, litCpy, shiftcount, flag
local restreg
restreg = reg
litCpy = literal
litCpy <<= 1
count = 0
shiftcount = 0
flag = 0
	if literal == 0
		clrf prod
		exitm
	endif
	movf reg, w
	bcf STATUS, C
	while (count < 8)
		litCpy <<= 1
		if (high litCpy)  == B'00000010'
			if flag == 0
				movwf prod
			else
				MULTL_B_B_SHIFT prod, shiftcount, restreg, 1
				addwf prod, f
			endif
			flag = 1
			litCpy = litCpy & 0x01FF
			shiftcount = 0
		endif
		if (high litCpy) == B'00000110'
			if flag == 0
				movwf prod
			else
				MULTL_B_B_SHIFT prod, (shiftcount - 1), restreg, 1
				addwf prod, f
			endif
			flag = 1
			MULTL_B_B_SHIFT prod, 1, reg, 1
			addwf prod, f
			litCpy = litCpy & 0x01FF
			shiftcount = 0
		endif
		if (high litCpy) == B'00000111'
			if flag == 0
				movwf prod
			else
				MULTL_B_B_SHIFT prod, (shiftcount - 2), restreg, 1
				addwf prod, f
			endif
			flag = 1
			shiftcount = 2			
		endif
		if ((high litCpy) & 0x0F) == B'00001110'
			MULTL_B_B_SHIFT prod, shiftcount, restreg, 1
			subwf prod, f
			if count < 7 
				bcf STATUS, C
			endif
			litCpy = litCpy & 0x01FF
			shiftcount = 0			
		endif
		if flag == 1
			shiftcount++
		endif
		count++
	endw
	MULTL_B_B_SHIFT prod, shiftcount - 1, restreg, 0
endm

;*Multiply reg by a literal and store a two byte result in prodH, prodL.  Generates fairly optimized code (24.5 instruction words/execution cycles on average).  Affects STATUS and WREG.
;*literal: constant to multiply reg by
;*reg: register to multiply by literal
;*prodH: register to retain high byte of the product
;*prodL: register to retain low byte of the product
MULTL_B_I macro literal, reg, prodH, prodL	;public
variable count, litCpy, shiftcount, flag
local restreg
restreg = reg
litCpy = literal
litCpy <<= 1
count = 0
shiftcount = 0
flag = 0
	if literal == 0
		clrf prodH
		clrf prodL
		exitm
	endif
	while (count < 8)
		litCpy <<= 1
		if (high litCpy)  == B'00000010'
			if flag == 0
				flag = 1
			else
				if flag == 1
					MULTL_B_I_SHIFT prodH, prodL, shiftcount, restreg, 1, 0
					flag = 2
				else
					MULTL_B_I_SHIFT prodH, prodL, shiftcount, restreg, 1, 1				
				endif
				addwf prodL, f
				btfsc STATUS, C
					incf prodH, f
				if count < 7				;added
					bcf STATUS, C			;added
				endif						;added
			endif
			litCpy = litCpy & 0x01FF
			shiftcount = 0
		endif
		if (high litCpy) == B'00000110'
			if flag == 0
				flag = 1
			else
				if flag == 1
					MULTL_B_I_SHIFT prodH, prodL, (shiftcount - 1), restreg, 1, 0
					flag = 2
				else
					MULTL_B_I_SHIFT prodH, prodL, (shiftcount - 1), restreg, 1, 1				
				endif
				addwf prodL, f
				btfsc STATUS, C
					incf prodH, f
				if count < 7				;added
					bcf STATUS, C			;added
				endif						;added
			endif
			if flag == 1
				MULTL_B_I_SHIFT prodH, prodL, 1, restreg, 1, 0				
			else	
				MULTL_B_I_SHIFT prodH, prodL, 1, restreg, 1, 1				
			endif
			addwf prodL, f
			btfsc STATUS, C
				incf prodH, f
			if count < 7				;added
				bcf STATUS, C			;added
			endif						;added
			litCpy = litCpy & 0x01FF
			flag = 2
			shiftcount = 0
		endif
		if (high litCpy) == B'00000111'
			if flag == 0
				flag = 1
			else
				if flag == 1
					MULTL_B_I_SHIFT prodH, prodL, (shiftcount - 2), restreg, 1, 0
					flag = 2
				else
					MULTL_B_I_SHIFT prodH, prodL, (shiftcount - 2), restreg, 1, 1				
				endif
				addwf prodL, f
				btfsc STATUS, C
					incf prodH, f
				if count < 7				;added
					bcf STATUS, C			;added
				endif						;added
			endif
			shiftcount = 3			
		endif
		if ((high litCpy) & 0x0F) == B'00001110'
			if flag == 1
				MULTL_B_I_SHIFT prodH, prodL, (shiftcount - 1), restreg, 1, 0
				flag = 2
			else
				MULTL_B_I_SHIFT prodH, prodL, (shiftcount - 1), restreg, 1, 1				
			endif
			subwf prodL, f
			btfss STATUS, C
				decf prodH, f
			if count < 7				;added
				bcf STATUS, C			;added
			endif						;added
			litCpy = litCpy & 0x01FF
			shiftcount = 0			
		endif
		if flag != 0
			shiftcount++
		endif
		count++
	endw
	if flag == 1
		MULTL_B_I_SHIFT prodH, prodL, (shiftcount - 1), restreg, 0, 0
	else
		MULTL_B_I_SHIFT prodH, prodL, (shiftcount - 1), restreg, 0, 1
	endif
endm

